#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include "gen.h"
#include "bitv.h"
#include "popu_t.h"
#include "set.h"
#include "graph_am.h"

/********* CONSTANT DEFINITIONS *******/
#define MAX_TRIALS             5000
#define MAX_NO_CHANGE_TRIALS   50000
#define INITIAL_PHEROMONE      10.0
#define PHEROMONE_PERSIST      0.995
#define PHEROMONE_MAX          7.0
#define PHEROMONE_MIN          0.08
#define PROB_FILTER            0.6
#define MAX_NODE_THRESH        FALSE
#define ALPHA                  2.0
#define BETA                   4.0
#define Q0                     0.0
#define ETA                    TRUE
#define PP_ITER                250
#define HEUR_ANTS              0

/******* DATA TYPE DEFINITIONS *****/
typedef struct roulette_s {
  float       f;
  int         id;
} roulette_t;

/******* GLOBAL VARIABLES  *******/
static Graph           g;
static population_t    pop;
static popu_member_t   bestMember;
static int             bestFitness;
static roulette_t      R[N];

/************ Extern Function prototypes **********/
extern int parse_udg_cap(FILE *fp, Graph *g, int var);


/************** FUNCTION PROTOTYPES ***************/
static void initializePheromone(Graph *g);
static void aco_capmds(Graph *g, population_t *pop, int nAnts, int nIter);
static void updatePheromone(Graph *g, bitvec_t *bvec, int f, int bF);
static int selectRandomNode(Graph *g, float totalPheromone, int maxNode);
static float calcTotalPheromoneFactor(Graph *g, int *maxNode);
static void minimizeDS(bitvec_t *bvec, int *fitness);
static void delNodefromDS(bitvec_t *bvec, int *fitness, int nodeId);
static void addNodetoDS(Graph *g, bitvec_t *bvec, int nodeId, int *fitness);
static void printNodesCov(Graph *g, bitvec_t *bvec);
static void heurCapMDS(Graph *g, bitvec_t *bvec, int *fitness);

int main(int argc, char **argv)
{
  FILE           *fp = NULL;
  int             nAnts = 0, nIter = 0;
  int             rc;
  unsigned long   numSecs;
  struct timeval  t1, t2;

  if (argc < 5) {
    fprintf(stderr, "Usage: aco-capmds <filename> (U)ni-cap/(V)ar-cap <numAnts> <maxGen>\n");
    exit(1);
  }

  if ((fp = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "Error opening file %s\n", argv[1]);
    exit(2);
  }

  /* Initialize global variables */
  initGraph(&g);

  if (argv[2][0] == 'u' || argv[2][0] == 'U')
    rc = parse_udg_cap(fp, &g, 0);
  else if (argv[2][0] == 'v' || argv[2][0] == 'V')
    rc = parse_udg_cap(fp, &g, 1);
  else {
    fprintf(stderr, "Error in file format parameter\n");
    exit(3);
  }
 
  if (rc == -1) {
    fprintf(stderr, "Error parsing the input file\n");
    exit(4);
  }

  nAnts = atoi(argv[3]);
  if (nAnts > 100) {
    fprintf(stderr, "Error: maximum number of ants can be 100\n");
    exit(5);
  }
  nIter = atoi(argv[4]);

  bestFitness = g.nodes;

/*
  printGraph(&g);
*/

  gettimeofday(&t1, NULL);

  /*** Call the ACO algorithm ***/
  aco_capmds(&g, &pop, nAnts, nIter);

  gettimeofday(&t2, NULL);

  numSecs = t2.tv_sec - t1.tv_sec;
  printf("%s %d %ld\n", argv[1], bestFitness, numSecs);
/*
  printf("Best Member = \n");
  bitPrint(&bestMember.vec);
*/

  exit(0);
}

/* We use the 'weight' parameter in the graph for storing pheromone    */
/* value of a node. This value is updated over time using the standard */
/* pheromone updation rules of evaporation and augmentation.           */
static void initializePheromone(Graph *g)
{
  int               i;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (!isValidNode(i, &(g->nodeTable[i])))
      continue;

    g->nodeTable[i].weight = INITIAL_PHEROMONE;
  }

  return;
}

static float calcTotalPheromoneFactor(Graph *g, int *maxNode)
{
  int         n, deg;
  float       total = 0.0, max = 0.0;

  for (n = 0; n <= g->maxNodeId; n++) {
    if (!isValidNode(n, &(g->nodeTable[n])))
      continue;

    if (ETA) {
      deg = calcWhiteDegCovbyCap(g, &(g->nodeTable[n]));
      g->nodeTable[n].pFact = ((pow(g->nodeTable[n].weight, ALPHA)) *
                               (pow(deg, BETA)));
      if (max < g->nodeTable[n].pFact) {
        max = g->nodeTable[n].pFact;
        *maxNode = n;
      }
      total += g->nodeTable[n].pFact;
    } else {
      total += g->nodeTable[n].weight;
      if (max < g->nodeTable[n].weight) {
        max = g->nodeTable[n].weight;
        *maxNode = n;
      }
    }
  }

  return total;
}

static void printNodesCov(Graph *g, bitvec_t *bvec)
{
  int         i, j;

  for (i = 0; i < bvec->members; i++) {
    if (bvec->vec[i] == 0)
      continue;

    printf("Node %d covers: ", i);
    for (j = 0; j < g->nodeTable[i].nodesCov.size; j++) 
      printf("%d ", g->nodeTable[i].nodesCov.set[j]);
    printf("\n");
  }

  return;
}

static int findMaxNode(Graph *g)
{
  int         n, maxNode = -1;
  float       max = 0.0;

  for (n = 0; n <= g->maxNodeId; n++) {
    if (!isValidNode(n, &(g->nodeTable[n])))
      continue;
    if (g->nodeTable[n].color == BLACK)
      continue;

    if (ETA && (max < g->nodeTable[n].pFact)) {
        max = g->nodeTable[n].pFact;
        maxNode = n;
    } else if (!ETA && (max < g->nodeTable[n].weight)) {
        max = g->nodeTable[n].weight;
        maxNode = n;
    }
  }

  return maxNode;
}

static void aco_capmds(Graph *g, population_t *pop, int nAnts, int nIter)
{
  int               i, j, nodeId, rc, maxNode = -1;
  int               fitness = 0, noChange = 0, dup = 0;
  float             totalPheromone = 0.0, pr;
  popu_member_t    *member = NULL;
  bitvec_t          bvec;
/*
  time_t            curtime;
  srand(time(&curtime));
*/

  initializePheromone(g);

  /* Preprocessing step to set pheromone values of nodes that contribute */
  /* to DS when using the heuristic higher than other nodes.             */
  for (i = 0; i < PP_ITER; i++) {
    fitness = 0;
    bitInit(&bvec, g->nodes);
    heurCapMDS(g, &bvec, &fitness);
    /* Update pheromone of the nodes in the solution found */
#if 0
    for (j = 0; j < bvec.members; j++) {
      if (bvec.vec[i] == 0)
        continue;
      g->nodeTable[j].weight += (1.0 / (float) fitness);
    }
#endif
    if (fitness < bestFitness)
      bestFitness = fitness;
    updatePheromone(g, &bvec, fitness, bestFitness);
  }

  for (i = 0; i < nIter; i++) {
    /* Initialize the population, i.e., the Ants information */
    popuInit(pop);

/*
    printf("[%d]:Heur fitness = ", i);
*/
    for (j = 0; j < HEUR_ANTS; j++) {
      /* Add HEUR_ANTS solutions as heuristic solutions          */
      bitInit(&bvec, g->nodes);
      fitness = 0;
      heurCapMDS(g, &bvec, &fitness);
      popuAdd(pop, &bvec, fitness);
/*
      printf("%d ", fitness);
*/
    }

    /* Repeat the following for every ANT:                            */
    /* While the nodes selected do not form a dominating set, DO:     */
    /* Based on the roulette wheel, select a node from nodes          */
    /* not yet included in the ANT and add it to the random walk      */
    /* of the ANT. Update total pheromone value and select next node  */
    for (j = 0; j < (nAnts - HEUR_ANTS); j++) {
/*
      printf("******* ITER = %d, ANT = %d ********\n", i, j);
*/
      /* Initialize the bit vector, color the graph WHITE and         */
      /* calculate total pheromone value of the nodes as of now       */
      bitInit(&bvec, g->nodes);
      initColorWhite(g);
      fitness = 0;
      totalPheromone = calcTotalPheromoneFactor(g, &maxNode);
/*
      printf("Total pheromone = %f\n", totalPheromone);
*/

      /* Repeat while the random walk is not yet a DS  */
      while (g->whiteNodes > 0) {
        pr =  rand() / (RAND_MAX + 1.0);
        if (MAX_NODE_THRESH && (pr < Q0)) {
          /* We need to find maxNode for every step if pr < Q0       */
          maxNode = findMaxNode(g);
          addNodetoDS(g, &bvec, maxNode, &fitness);
          if (ETA)
            totalPheromone -= g->nodeTable[maxNode].pFact;
          else
            totalPheromone -= g->nodeTable[maxNode].weight;
        } else {
          nodeId = selectRandomNode(g, totalPheromone, maxNode);
/*
          printf("Node to be added to the DS = %d, white nodes = %d\n", 
                nodeId, g->whiteNodes);
          printf("[I(%d):A(%d):Total Pheromone = %f\n", i, j, totalPheromone);
*/
          addNodetoDS(g, &bvec, nodeId, &fitness);
          if (ETA)
            totalPheromone -= g->nodeTable[maxNode].pFact;
          else
            totalPheromone -= g->nodeTable[nodeId].weight;
        }
      }

/*
      printf("DS %d selected = \n", bvec.count);
      bitPrint(&bvec);
      printNodesCov(g, &bvec);
*/
      minimizeDS(&bvec, &fitness);

      if ((rc = popuAdd(pop, &bvec, fitness)) < 0) {
        dup++;
      } 
    }

    fitness = popuBestMember(pop, &member);

/*
    printf("  Best fitness = %d\n", fitness);
    printf("%d %d\n", i, fitness); 
    printf("Fitness of the current Ant = %d, best Ant = %d\n", fitness,
            bestFitness);
    printf("Best member found in iteration %d = \n", i);
    bitPrint(&(member->vec));
*/

    if (fitness < bestFitness) {
      noChange = 0;
      bestFitness = fitness;
      memcpy(&bestMember, member, sizeof(popu_member_t));
    } else {
      noChange++;
      if (noChange == MAX_TRIALS)
        break;
    }
    updatePheromone(g, &(member->vec), fitness, bestFitness);
/*
    printf("i:%d, fitness:%d\n", i, fitness);
*/
  }

/*
  printf("iterations = %d, dup = %d\n", i, dup);
*/
  return;
}

static int selectRandomNode(Graph *g, float totalPheromone, int maxNode)
{
  float            sum = 0.0, selsum = 0.0;
  double           myrand;
  int              i, index = 0, retID;

  memset(&R, 0, sizeof(roulette_t) * N);
/*
  printf("Total pheromone = %f\n", totalPheromone);
*/

  for (i = 0; i <= g->maxNodeId; i++) {
    if (!isValidNode(i, &(g->nodeTable[i])))
      continue;

    /* If the node is not BLACK, i.e., not included in the random     */
    /* walk of the ANT, we compute its probability of being selected  */
    if (g->nodeTable[i].color != BLACK) {
/*
      printf("Node %d's pheromone = %f\n", i, g->nodeTable[i].weight);
*/
      if (ETA)
        R[index].f = g->nodeTable[i].pFact / totalPheromone;
      else
        R[index].f = g->nodeTable[i].weight / totalPheromone;
      R[index].id = i;
      sum += R[index].f;
      index++;
    }
  }

/*
  printf("Index = %d, sum = %f, Array before sorting = \n", index, sum);
  for (i = 0; i < index; i++) {
    printf("R[%d].f = %f R[%d].id = %d\n", i, R[i].f, i, R[i].id);
  }
*/

  myrand = rand() / (RAND_MAX + 1.0);
  for (i = 0; i < index; i++) {
    selsum += R[i].f;
    if (selsum >= myrand)
      break;
  }

  if (i < index)
    retID = R[i].id;
  else
    retID = R[index - 1].id;

/*
  printf("Myrand = %f, Selsum = %f, retID %d\n", myrand, selsum, retID);
*/

  return retID;
}

static void updatePheromone(Graph *g, bitvec_t *bvec, int f, int bF)
{
  int            i;

/*
  printf("Updating pheromone values f = %d, bF = %d...\n", f, bF);
*/

  for (i = 0; i <= g->maxNodeId; i++) {
    if (!isValidNode(i, &(g->nodeTable[i])))
      continue;

    /* If the nodeId is present in the bit vector, it is part of DS     */
    /* Hence, we update its pheromone by evaporating as well as         */
    /* augmenting its value. Otherwise, we just evaporate the pheromone */
    if (bvec->vec[i] == 1) {
      g->nodeTable[i].weight = (g->nodeTable[i].weight * PHEROMONE_PERSIST) + 
                               (2.0 / (1.0 + (float) f - (float) bF));
    } else {
      g->nodeTable[i].weight = (g->nodeTable[i].weight * PHEROMONE_PERSIST);
    }
    if (g->nodeTable[i].weight < PHEROMONE_MIN) {
      g->nodeTable[i].weight = PHEROMONE_MIN;
    }
/*
    else if (g->nodeTable[i].weight > PHEROMONE_MAX) {
      printf("Pheromone %f exceeds MAX value\n", g->nodeTable[i].weight);
      g->nodeTable[i].weight = PHEROMONE_MAX;
    }
*/

/*
    printf("%s:%d pheromone = %f\n", __FUNCTION__, i, g->nodeTable[i].weight);
*/
  }

  return;
}

static void addNodetoDS(Graph *g, bitvec_t *bvec, int nodeId, int *fitness)
{
  bitSet(bvec, nodeId);
  if (g->nodeTable[nodeId].color == WHITE) {
    g->whiteNodes--;
/*
    g->nodeTable[nodeId].resCap--;
*/
  }

  g->nodeTable[nodeId].ref++;
  addNodetoNodeSet(nodeId, &(g->nodeTable[nodeId].nodesCov));
  g->nodeTable[nodeId].color = BLACK;
  markCapAdjNodesGrey(g, nodeId, TRUE);
  (*fitness)++;

  return;
}


static void delNodefromDS(bitvec_t *bvec, int *fitness, int nodeId)
{
  int            i, nid;

  bitClear(bvec, nodeId);
  g.nodeTable[nodeId].color = GREY;

#if 0
  AP: We should not clear the 'ref' of 'nodeId' in capmds as it may
      or may not cover itself. We do it only for 'nodesCov' below.

  /* If bit is being cleared as part of mutation, it may not be   */
  /* a redundant node. Hence, when we decrement ref, it may be 0  */
  /* In such a case, the color should be changed to WHITE. If     */
  /* bit is being cleared from minimizeDS, it is a redundant node */
  /* So, the color should be changed to GREY as some other node   */
  /* will be referring to it.                                     */
  g.nodeTable[nodeId].ref--;
  if (g.nodeTable[nodeId].ref == 0) {
    g.nodeTable[nodeId].color = WHITE;
    g.whiteNodes++;
  } else {
    g.nodeTable[nodeId].color = GREY;
  }
#endif /* 0 */

  /* Now, update the reference of the adj. nodes covered by this  */
  /* node and change their color accordingly. As earlier, for     */
  /* mutation the color may become WHITE whereas in minimization, */
  /* they will remain GREY                                        */
/*
  printf("%s: Decrementing refcount of covered nodes of %d\n", __FUNCTION__,
          nodeId);
*/
  for (i = 0; i < g.nodeTable[nodeId].nodesCov.size; i++) {
    nid = g.nodeTable[nodeId].nodesCov.set[i];
/*
    printf("%d ", nid);
*/
    g.nodeTable[nid].ref--;
    if (g.nodeTable[nid].ref == 0) {
      g.nodeTable[nid].color = WHITE;
      g.whiteNodes++;
    } else {
      g.nodeTable[nid].color = GREY;
    }
  }
/*
  printf("\n");
*/

  /* Update the fitness value and return                          */
  (*fitness)--;

  return;
}

static void minimizeDS(bitvec_t *bvec, int *fitness)
{
  int               i, index; 
  int               maxCap, cap, capDeg, minDeg, deg;
  int               minId, nid;
  nodeSet           redSet;
  float             myrand;

/*
  printf("In Function %s, set input\n", __FUNCTION__);
  bitPrint(bvec);
  printNodesCov(&g, bvec);
*/

  /* Initialize and collect the set of nodes which are redundant   */
  /* in the given dominating set.                                  */
  while (TRUE) {
    initNodeSet(&redSet);
    for (i = 0; i < bvec->members; i++) {
      if (bvec->vec[i] == 0)
        continue;
      if (isNodeDSRedundant(&g, bvec, i)) {
/*
        printf("adding node %d to red set\n", i);
*/
        addNodetoNodeSet(i, &redSet);
      }
    }

    if (redSet.size == 0) {
      break;
    }

/*
    printf("%s: Redundant set of size %d is found\n", __FUNCTION__, 
           redSet.size);
    printNodeSet(&redSet);
*/

    myrand = ((float) rand()) / ((float) (RAND_MAX + 1.0));
    if (myrand <= PROB_FILTER) {
      index = rand() % redSet.size;
/*
      printf("Clearing redundant node %d from the bit vector\n", 
              redSet.set[index]);
*/
      delNodefromDS(bvec, fitness, redSet.set[index]);
    } else {
      minDeg = N;
      capDeg = N;
      maxCap = -1;
      nid = -1;
      minId = -1;
      for (i = 0; i < redSet.size; i++) {
        cap = g.nodeTable[redSet.set[i]].resCap;
        deg = g.nodeTable[redSet.set[i]].degree;
        if (cap > 0) {
          if (maxCap < cap) {
            maxCap = cap;
            capDeg = deg;
            nid = redSet.set[i];
          } else if ((maxCap == cap) && (deg < capDeg)) {
            capDeg = deg;
            nid = redSet.set[i];
          }
        } else if (minDeg > deg) {
          minDeg = deg;
          minId = redSet.set[i];
        }
      }

      if (nid != -1) {
/*
        printf("Node %d with resCap = %d found in redSet, removing it\n",
                nid, maxCap);
*/
        delNodefromDS(bvec, fitness, nid);
      } else {
/*
        printf("No node with resCap, removing %d with degree %d from redSet\n",
                minId, minDeg);
*/
        delNodefromDS(bvec, fitness, minId);
      }
    }
  }

  return;
}

static void heurCapMDS(Graph *g, bitvec_t *bvec, int *fitness)
{
  int           i, maxNode = -1, maxCap = 0, cap;
  Node_t       *node;
  boolean       useAllCap = FALSE;

  /* Initialize the color of all nodes to WHITE */
  initColorWhite(g);

  while (TRUE) {
    maxCap = 0;
    cap = 0;

    for (i = 0; i <= g->maxNodeId; i++) {
      if (!isValidNode(i, &(g->nodeTable[i])))
        continue;

      node = &(g->nodeTable[i]); 
      if (node->color != BLACK) {
        cap = calcWhiteDegCovbyCap(g, node);
/*
        printf("cap returned for %d = %d\n", g->nodeTable[i].nodeId, cap);
*/
        if (maxCap < cap) {
          maxCap = cap;
          maxNode = node->nodeId;
/*
          printf("Node with max white degree covered by capacity (%d) is %d\n",
                  cap, maxNode);
*/
        } else if (maxCap == cap) {
          if (g->nodeTable[i].degree < g->nodeTable[maxNode].degree) {
            maxCap = cap;
            maxNode = node->nodeId;
          }
        }
      }
    }
    /* If no node with white degree is found, then, no node is white */
    if (maxCap == 0) {
/*
      printf("No node with white degree is found\n");
*/
      break;
    }

    /* Add the maxNode to S and mark it black; mark all its neighbors GREY */
/*
    printf("Adding node %d with capacity %d to the DS\n", maxNode, maxCap);
    if (g->nodeTable[maxNode].color == WHITE)
      g->nodeTable[maxNode].resCap--;
*/
    g->nodeTable[maxNode].color = BLACK;
    g->nodeTable[maxNode].ref++;
    bitSet(bvec, maxNode);
    (*fitness)++;
    markCapAdjNodesGrey(g, maxNode, useAllCap);
  }

  return;
}

#if 0
static void minimizeDS(bitvec_t *bvec, int *fitness)
{
  int               i, index;
  nodeSet           redSet;
/*
  int               minDeg, deg, minId;
  float             myrand;
*/

/*
  printf("In Function %s, set input\n", __FUNCTION__);
  bitPrint(bvec);
*/

  /* Initialize and collect the set of nodes which are redundant   */
  /* in the given dominating set.                                  */
  while (TRUE) {
    initNodeSet(&redSet);
    for (i = 0; i < bvec->members; i++) {
      if (bvec->vec[i] == 0)
        continue;
/*
      printf("Checking if node %d is redundant\n", i);
*/
      if (isNodeDSRedundant(&g, bvec, i)) {
/*
        printf("adding node %d to red set\n", i);
*/
        addNodetoNodeSet(i, &redSet);
      }
    }

    if (redSet.size == 0) {
/*
      printf("No more redundant nodes in the member\n");
*/
      break;
    }

/*
    printf("%s: Redundant set of size %d is found\n", __FUNCTION__, 
           redSet.size);
    printNodeSet(&redSet);

    myrand = rand() / (RAND_MAX + 1.0);
    if (myrand <= PROB_FILTER) {
*/
      index =  rand() % redSet.size;
/*
      printf("Clearing redundant node %d from the bit vector\n", 
              redSet.set[index]);
*/
      delNodefromDS(bvec, fitness, redSet.set[index]);

#if 0
    }
 else {
      minDeg = N;
      minId = -1;
      for (i = 0; i < redSet.size; i++) {
        deg = g.nodeTable[redSet.set[i]].degree;
        if (minDeg > deg) {
          minDeg = deg;
          minId = redSet.set[i];
        }
      }

      delNodefromDS(bvec, fitness, minId);
    }
#endif /* 0 */
  }

  return;
}
#endif /* 0 */
