#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "gen.h"
#include "bitv.h"
#include "set.h"
#include "domset.h"
#include "graph_am.h"

typedef struct nodeDeg_s {
  int     nodeId;
  int     degree;
} nodeDeg_t;


/* Function to initialize the graph */
void initGraph(Graph *g)
{
  g->nodes = 0;
  g->whiteNodes = 0;
  g->maxNodeId = -1;
  memset(g->nodeTable, 0, sizeof(Node_t) * N);
  g->nodeTable[0].nodeId = -1;
  memset(g->adjMatrix, 0, (sizeof(char) * (N * N)));
  memset(g->neighborIds, 0, (sizeof(int) * (N * N)));
}

/* Function to determine if an index of the vector of the adjacency */
/* list has a valid node of the graph.                              */
boolean isValidNode(int index, Node_t *node)
{
  if ((index == 0) && (node->nodeId == -1))
    return FALSE;
  else if ((index != 0) && (node->nodeId == 0))
    return FALSE;
  else
    return TRUE;
}

/* Function to initialize the color of every node in the given     */
/* to WHITE. Used in finding MDS/MIS.                              */
void initColorWhite(Graph *g)
{
  int        i;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (isValidNode(i, &(g->nodeTable[i]))) {
      g->nodeTable[i].color  = WHITE;
      g->nodeTable[i].ref    = 0;
      g->nodeTable[i].resCap = g->nodeTable[i].capacity;
      initNodeSet(&(g->nodeTable[i].nodesCov));
    }
  }
  g->whiteNodes = g->nodes;

  return;
}

void printNodeColors(Graph *g)
{
  int           i;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (isValidNode(i, &(g->nodeTable[i])))  
      printf("%d:%d ", i, g->nodeTable[i].color);
  }
  printf("\n");
}

void printWhiteNodes(Graph *g)
{
  int           i;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (isValidNode(i, &(g->nodeTable[i])) && (g->nodeTable[i].color == WHITE))
      printf("%d ", i);
  }
  printf("\n");
}

/* Function to print the adjacency list of a given node */
static void printAdjList(Graph *g, int id)
{
  int            i;

  for (i = 0; i < g->nodeTable[id].degree; i++) {
      printf("%d\t", g->neighborIds[id][i]);
  }

  return;
}

/* Function to print the graph as an adjacency list */
void printGraph(Graph *g)
{
  int        i;

  printf("Maximum Node ID in the graph = %d\n", g->maxNodeId);
  printf("Number of nodes in the graph = %d\n", g->nodes);

  if (g->nodeTable[0].nodeId != -1) {
    printf("Node %d[%d,%d,%d,%f]:\t", g->nodeTable[0].nodeId, 
           g->nodeTable[0].degree, g->nodeTable[0].capacity, 
           g->nodeTable[0].resCap, g->nodeTable[0].weight);
    printAdjList(g, 0);
  }

  printf("\n");
  for (i = 1; i <= g->maxNodeId; i++) {
    if (g->nodeTable[i].nodeId == 0)
      continue;

    printf("Node %d[%d,%d,%d,%f]:\t", g->nodeTable[i].nodeId, 
            g->nodeTable[i].degree, g->nodeTable[i].capacity,
            g->nodeTable[i].resCap, g->nodeTable[i].weight);
    printAdjList(g, i);
    printf("\n");
  }

  return;
}

int calcWhiteDegree(Graph *g, Node_t *node)
{
  int           wDeg = 0, i, nodeId;

  if (node->color == WHITE)
    wDeg++;

  nodeId = node->nodeId;
  for (i = 0; i <= g->maxNodeId; i++) {
    if (g->adjMatrix[nodeId][i] == 0)
      continue;
    else if (g->nodeTable[i].color == WHITE)
      wDeg++;
  }

  return wDeg;
}

int calcWhiteDegCovbyCap(Graph *g, Node_t *node)
{
  int           wDeg = 0, i, nodeId;

  if (node->capacity == 0)
    return wDeg;

  if (node->color == WHITE)
    wDeg++;
  if (wDeg == node->capacity)
    return wDeg;

  nodeId = node->nodeId;
  for (i = 0; i <= g->maxNodeId; i++) {
    if (g->adjMatrix[nodeId][i] == 0)
      continue;

    if (g->nodeTable[i].color == WHITE) {
      wDeg++;
      if (wDeg == node->capacity)
        return wDeg;
    }
  }

  return wDeg;
}

static int degcmp(const void *n1, const void *n2)
{
  nodeDeg_t      *nid1 = (nodeDeg_t *) n1, *nid2 = (nodeDeg_t *) n2;

  if (nid1->degree < nid2->degree)
    return -1;
  else if (nid1->degree > nid2->degree)
    return 1;
  else
    return 0;
}

void markLdegAdjNodesGrey(Graph *g, int nodeId)
{
  int            i, neighbor, index = 0;
  nodeDeg_t      wNodes[N];

  if (!isValidNode(nodeId, &(g->nodeTable[nodeId]))) {
    fprintf(stderr, "%s: Unexpected invalid node ID %d\n", 
            __FUNCTION__, nodeId);
    return;
  }

  if (g->nodeTable[nodeId].resCap == 0)
    return;

  for (i = 0; i < g->nodeTable[nodeId].degree; i++) {
    neighbor = g->neighborIds[nodeId][i];
    if (g->nodeTable[neighbor].color == WHITE) {
      wNodes[index].nodeId = neighbor;
      wNodes[index].degree = g->nodeTable[neighbor].degree;
      index++;
    }
  }

/*
  printf("White nodes before sorting\n");
  for (i = 0; i < index; i++)
    printf("Node[%d]:%d\n", wNodes[i].nodeId, wNodes[i].degree);
*/

  qsort(wNodes, index, sizeof(nodeDeg_t), degcmp);
 
/*
  printf("White nodes after sorting\n");
  for (i = 0; i < index; i++)
    printf("Node[%d]:%d\n", wNodes[i].nodeId, wNodes[i].degree);
*/

  for (i = 0; i < index; i++) {
    neighbor = wNodes[i].nodeId;
    g->nodeTable[neighbor].color = GREY;
    addNodetoNodeSet(neighbor, &(g->nodeTable[nodeId].nodesCov));
    g->whiteNodes--;
    g->nodeTable[neighbor].ref++;
    g->nodeTable[nodeId].resCap--;

    if (g->nodeTable[nodeId].resCap == 0)
      return;
  }

  return;
}

void markAdjNodesGrey(Graph *g, int nodeId, boolean cap)
{
  int            i, neighbor;

  if (!isValidNode(nodeId, &(g->nodeTable[nodeId]))) {
    fprintf(stderr, "%s: Unexpected invalid node ID %d\n", 
            __FUNCTION__, nodeId);
    return;
  }

  if (cap) {
    if (g->nodeTable[nodeId].resCap == 0)
      return;
  }

/*
  printf("Marking neighbors of %d GREY : ", nodeId);
*/
  for (i = 0; i < g->nodeTable[nodeId].degree; i++) {
    neighbor = g->neighborIds[nodeId][i];
    if (g->nodeTable[neighbor].color == WHITE) {
/*
      printf("%d ", neighbor);
*/
      g->nodeTable[neighbor].color = GREY;
      addNodetoNodeSet(neighbor, &(g->nodeTable[nodeId].nodesCov));
      g->whiteNodes--;
      g->nodeTable[neighbor].ref++;
      if (cap) {
        g->nodeTable[nodeId].resCap--;
      }
    }

    if (cap) {
      if (g->nodeTable[nodeId].resCap == 0) {
/*
        printf("\n");
*/
        return;
      }
    }
  }

/*
  printf("\n");
*/
  return;
}

#if 0
void markCapAdjNodesGrey(Graph *g, int nodeId, boolean useAllCap)
{
  int            i, neighbor;
  int            index;
  nodeSet        wNodes, gbNodes;

  if (!isValidNode(nodeId, &(g->nodeTable[nodeId]))) {
    fprintf(stderr, "%s: Unexpected invalid node ID %d\n", 
            __FUNCTION__, nodeId);
    return;
  }

  if (g->nodeTable[nodeId].resCap == 0)
    return;

  initNodeSet(&wNodes);
  initNodeSet(&gbNodes);

  /* 'nodeId' was already GREY and so it was not included in covered  */
  /* nodes. It should be in gbNodes so that it covers itself if there */
  /* is excess capacity.                                              */
  if (!isSetMember(nodeId, &(g->nodeTable[nodeId].nodesCov))) {
/*
    printf("Adding nodeId %d to gbNodes as it was already GREY\n", nodeId);
*/
    addNodetoNodeSet(nodeId, &gbNodes);
  }

  /* Construct the 'wNodes' and 'gbNodes' sets by going through the  */
  /* neighbors of the given node 'nodeId'.                           */
  for (i = 0; i < g->nodeTable[nodeId].degree; i++) {
    neighbor = g->neighborIds[nodeId][i];
/*
    printf("Neighbor %d of node %d is %s\n", neighbor, nodeId,
            (g->nodeTable[neighbor].color == WHITE) ? "WHITE" : "NON-WHITE");
*/
    if (g->nodeTable[neighbor].color == WHITE)
      addNodetoNodeSet(neighbor, &wNodes);
    else
      addNodetoNodeSet(neighbor, &gbNodes);
  }

/*
  printf("White neighbors of %d are \n", nodeId);
  printNodeSet(&wNodes);
*/

  /* Loop through the WHITE neighbors until all are covered or       */
  /* residual capacity of the node runs out, whichever is earlier :) */
  index = 0;
  while (index < wNodes.size) {
    neighbor = wNodes.set[index];
/*
    printf("Neighbor %d being covered by %d with resCap=%d\n", neighbor,
            nodeId, g->nodeTable[nodeId].resCap);
*/
    g->whiteNodes--;
    g->nodeTable[neighbor].color = GREY;
    addNodetoNodeSet(neighbor, &(g->nodeTable[nodeId].nodesCov));
    g->nodeTable[neighbor].ref++;
    g->nodeTable[nodeId].resCap--;
    index++;

/*
    printf("Added node %d's neighbor %d to covered nodes [%d]\n", nodeId,
            neighbor, g->nodeTable[nodeId].nodesCov.size);
*/

    if (g->nodeTable[nodeId].resCap == 0) {
      return;
    }
  }

  /* When this function is used in heur_randcap.c, it does not care    */
  /* about redundant coverage. We just need to cover all WHITE nodes   */
  /* up to a given capacity to be covered. Hence, we return from here. */
  /* The rest of the code for redundant coverage is primarily for GA   */
  /* and ACO algorithms so they can minimize the DS cardinality.       */
  if (!useAllCap)
    return;

/*
  printf("Residual capacity = %d, numGB = %d\n", 
          g->nodeTable[nodeId].resCap, numGB);
*/

  index = 0;
  while ((g->nodeTable[nodeId].resCap != 0) && (index < gbNodes.size)) { 
    neighbor = gbNodes.set[index];
/*
    printf("Covering neighbor %d with rescap\n", neighbor);
*/
    addNodetoNodeSet(neighbor, &(g->nodeTable[nodeId].nodesCov));
    g->nodeTable[neighbor].ref++;
    g->nodeTable[nodeId].resCap--;
    index++;
  }

  return;
}
#endif 

void markCapAdjNodesGrey(Graph *g, int nodeId, boolean useAllCap)
{
  int            i, neighbor;
  int            index;
  nodeSet        wNodes, gbNodes;

  if (!isValidNode(nodeId, &(g->nodeTable[nodeId]))) {
    fprintf(stderr, "%s: Unexpected invalid node ID %d\n", 
            __FUNCTION__, nodeId);
    return;
  }

  if (g->nodeTable[nodeId].resCap == 0) {
/*
    printf("NodeId %d has no residual cap...so returning\n", nodeId);
*/
    return;
  }

  initNodeSet(&wNodes);
  initNodeSet(&gbNodes);

  /* 'nodeId' was already GREY and so it was not included in covered  */
  /* nodes. It should be in gbNodes so that it covers itself if there */
  /* is excess capacity.                                              */
  if (!isSetMember(nodeId, &(g->nodeTable[nodeId].nodesCov))) {
/*
    printf("Adding nodeId %d to gbNodes as it was already GREY\n", nodeId);
*/
    addNodetoNodeSet(nodeId, &gbNodes);
  }

  /* Construct the 'wNodes' and 'gbNodes' sets by going through the  */
  /* neighbors of the given node 'nodeId'.                           */
  for (i = 0; i < g->nodeTable[nodeId].degree; i++) {
    neighbor = g->neighborIds[nodeId][i];
/*
    printf("Neighbor %d of node %d is %s\n", neighbor, nodeId,
            (g->nodeTable[neighbor].color == WHITE) ? "WHITE" : "NON-WHITE");
*/
    if (g->nodeTable[neighbor].color == WHITE)
      addNodetoNodeSet(neighbor, &wNodes);
    else
      addNodetoNodeSet(neighbor, &gbNodes);
  }

  /* Loop through the WHITE neighbors until all are covered or       */
  /* residual capacity of the node runs out, whichever is earlier :) */
  while (wNodes.size > 0) {
    index = rand() % wNodes.size;
    neighbor = wNodes.set[index];
/*
    printf("Neighbor %d being covered by %d with resCap=%d\n", neighbor,
            nodeId, g->nodeTable[nodeId].resCap);
*/
    g->whiteNodes--;
    g->nodeTable[neighbor].color = GREY;
    addNodetoNodeSet(neighbor, &(g->nodeTable[nodeId].nodesCov));
    delNodefromNodeSet(neighbor, &wNodes);
    g->nodeTable[neighbor].ref++;
    g->nodeTable[nodeId].resCap--;

/*
    printf("Added node %d's neighbor %d to covered nodes [%d]\n", nodeId,
            neighbor, g->nodeTable[nodeId].nodesCov.size);
*/

    if (g->nodeTable[nodeId].resCap == 0) {
/*
      printf("Ran out of capacity of %d\n", nodeId);
*/
      return;
    }
  }

  /* When this function is used in heur_randcap.c, it does not care    */
  /* about redundant coverage. We just need to cover all WHITE nodes   */
  /* up to a given capacity to be covered. Hence, we return from here. */
  /* The rest of the code for redundant coverage is primarily for GA   */
  /* and ACO algorithms so they can minimize the DS cardinality.       */
  if (!useAllCap)
    return;

/*
  printf("Residual capacity = %d, numGB = %d\n", 
          g->nodeTable[nodeId].resCap, numGB);
*/

  while ((g->nodeTable[nodeId].resCap != 0) && (gbNodes.size != 0)) { 
    index = rand() % gbNodes.size;
    neighbor = gbNodes.set[index];
/*
    printf("Covering neighbor %d with rescap\n", neighbor);
*/
    addNodetoNodeSet(neighbor, &(g->nodeTable[nodeId].nodesCov));
    delNodefromNodeSet(neighbor, &gbNodes);
    g->nodeTable[neighbor].ref++;
    g->nodeTable[nodeId].resCap--;
  }

/*
  printf("%s: Done coverage of node %d\n", __FUNCTION__, nodeId);
*/
  return;
}

#if 0
void markAdjNodesGreyandAddDomSet(Graph *g, int nodeId, int *set, 
                                  int *domNodes)
{
  int            i;

  if (!isValidNode(nodeId, &(g->nodeTable[nodeId]))) {
    fprintf(stderr, "%s: Unexpected invalid node ID %d\n", 
            __FUNCTION__, nodeId);
    return;
  }

  *domNodes = 0;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (g->adjMatrix[nodeId][i] == 0)
      continue;

    /* Node with index i is a neighbor of nodeID and so the color/ref   */
    /* need to be updated. We stop when we reach the capacity of nodeID */
    if (g->nodeTable[i].color == WHITE) {
      g->nodeTable[i].color = GREY;
      set[*domNodes] = i;
      (*domNodes)++;
      g->whiteNodes--;
    }
#if 0
    /* Increment reference irrespective of whether the adjacent node */
    /* is WHITE, GREY or BLACK as it is covered again by this node   */
    g->nodeTable[i].ref++;
#endif /* 0 */
  }

  return;
}
#endif /* 0 */

void addWtNodetoGraph(Graph *g, int nodeId, float weight)
{
  g->nodes++;
  g->whiteNodes++;
  if (nodeId > g->maxNodeId)
    g->maxNodeId = nodeId;
  g->nodeTable[nodeId].nodeId = nodeId;
  g->nodeTable[nodeId].color = WHITE;
  g->nodeTable[nodeId].weight = weight;
  g->nodeTable[nodeId].degree = 0;
  g->nodeTable[nodeId].ref = 0;
}

void addCapNodetoGraph(Graph *g, int nodeId, int capacity)
{
  g->nodes++;
  g->whiteNodes++;
  if (nodeId > g->maxNodeId)
    g->maxNodeId = nodeId;
  g->nodeTable[nodeId].nodeId = nodeId;
  g->nodeTable[nodeId].color = WHITE;
  g->nodeTable[nodeId].capacity = capacity;
  g->nodeTable[nodeId].resCap = capacity;
  g->nodeTable[nodeId].degree = 0;
  g->nodeTable[nodeId].ref = 0;
}

/* Function to add a node to the graph */
static void addNodetoGraph(Graph *g, int nodeId)
{
  g->nodes++;
  g->whiteNodes++;
  if (nodeId > g->maxNodeId)
    g->maxNodeId = nodeId;
  g->nodeTable[nodeId].nodeId = nodeId;
  g->nodeTable[nodeId].color = WHITE;
  g->nodeTable[nodeId].degree = 0;
  g->nodeTable[nodeId].ref = 0;
  g->nodeTable[nodeId].capacity = 0.0;
  g->nodeTable[nodeId].resCap = 0.0;
}

static int addAdjListNode(Graph *g, int n1, int n2)
{
  /* Update the adjacency matrix */
  g->adjMatrix[n1][n2] = 1;

  /* Update the neighbor ID list */
  g->neighborIds[n1][g->nodeTable[n1].degree] = n2;
  (g->nodeTable[n1].degree)++;

  return 0;
}

/* Function to add an edge to the graph */
int addEdgetoGraph(Graph *g, int fromNode, int toNode)
{
  int             rc;

  /* A nodeId of 0 in the vector represents an uninitialized entry       */
  /* except for nodeId 0 for which a -1 represents uninitialized value.  */
  /* Add the node to the graph in such a case before adding the node to  */
  /* the adjacency list of the respective nodes.                         */
  if (((fromNode == 0) && (g->nodeTable[fromNode].nodeId == -1)) ||
      ((fromNode != 0) && (g->nodeTable[fromNode].nodeId == 0)))
  {
    addNodetoGraph(g, fromNode);
  }

  if (((toNode == 0) && (g->nodeTable[toNode].nodeId == -1)) ||
      ((toNode != 0) && (g->nodeTable[toNode].nodeId == 0)))
  {
    addNodetoGraph(g, toNode);
  }

  /* Update the adjacency lists of fromNode and toNode as this is an  */
  /* undirected graph.                                                */
  if ((rc = addAdjListNode(g, fromNode, toNode)) == -1)
    return rc;

  if ((rc = addAdjListNode(g, toNode, fromNode)) == -1)
    return rc;

  return 0;
}

/* Function to add a directed edge to the graph */
int addDirEdgetoGraph(Graph *g, int fromNode, int toNode)
{
  int             rc;

  /* A nodeId of 0 in the vector represents an uninitialized entry       */
  /* except for nodeId 0 for which a -1 represents uninitialized value.  */
  /* Add the node to the graph in such a case before adding the node to  */
  /* the adjacency list of the respective nodes.                         */
  if (((fromNode == 0) && (g->nodeTable[fromNode].nodeId == -1)) ||
      ((fromNode != 0) && (g->nodeTable[fromNode].nodeId == 0)))
  {
    addNodetoGraph(g, fromNode);
  }

  if (((toNode == 0) && (g->nodeTable[toNode].nodeId == -1)) ||
      ((toNode != 0) && (g->nodeTable[toNode].nodeId == 0)))
  {
    addNodetoGraph(g, toNode);
  }

  /* Update the adjacency list of only fromNode as this is a        */
  /* directed graph.                                                */
  if ((rc = addAdjListNode(g, fromNode, toNode)) == -1)
    return rc;

  return 0;
}

/* Function to add a directed edge to a graph with uniform node capacity */
int addDirEdgetoCapGraph(Graph *g, int fromNode, int toNode, int capacity)
{
  int             rc;

  /* A nodeId of 0 in the vector represents an uninitialized entry       */
  /* except for nodeId 0 for which a -1 represents uninitialized value.  */
  /* Add the node to the graph in such a case before adding the node to  */
  /* the adjacency list of the respective nodes.                         */
  if (((fromNode == 0) && (g->nodeTable[fromNode].nodeId == -1)) ||
      ((fromNode != 0) && (g->nodeTable[fromNode].nodeId == 0)))
  {
    addCapNodetoGraph(g, fromNode, capacity);
  }

  if (((toNode == 0) && (g->nodeTable[toNode].nodeId == -1)) ||
      ((toNode != 0) && (g->nodeTable[toNode].nodeId == 0)))
  {
    addCapNodetoGraph(g, toNode, capacity);
  }

  /* Update the adjacency list of only fromNode as this is a        */
  /* directed graph.                                                */
  if ((rc = addAdjListNode(g, fromNode, toNode)) == -1)
    return rc;

  return 0;
}

/* Function to construct the set for the nodes in the given graph */
void constructSet(Graph *g, nodeSet *V)
{
  int        i;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (isValidNode(i, &(g->nodeTable[i]))) 
      addNodetoNodeSet(i, V);
  }

  return;
}

/* Function to determine if node n2 is a neighbor of node n1 in graph G */
boolean isNeighbor(Graph *G, int n1, int n2)
{
  if (G->adjMatrix[n1][n2] == 1)
    return TRUE;
  else
    return FALSE;
}

/* Function to walk the adjacency list of a given node and */
/* add all its neighbors to the given set.                 */
static void addAdjListNodestoSet(Graph *g, int nodeId, nodeSet *D)
{
  int           i;

  for (i = 0; i < g->nodes; i++) {
    if (g->adjMatrix[nodeId][i] == 1) {
      addNodetoNodeSet(i, D);
      g->nodeTable[i].color = GREY;
      g->nodeTable[i].ref++;
      g->whiteNodes--;
    }
/*
    printf("Set color of %d to %d:%d\n", node->nodeId, GREY,
            g->adjList[node->nodeId].color);
*/
  }

  return;
}

void getWhiteNodeSet(Graph *g, nodeSet *S)
{
  int       i;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (!isValidNode(i, &(g->nodeTable[i])))
      continue;
    if (g->nodeTable[i].color == WHITE) {
      addNodetoNodeSet(g->nodeTable[i].nodeId, S);
    }
  }

  return;
}

int nodesCoveredbyVec(Graph *g, bitvec_t *bvec)
{
  int             i, count = 0;

  for (i = 0; i < bvec->members; i++) {
    if (bvec->vec[i] == 1) {
      g->nodeTable[i].color = BLACK;
      g->nodeTable[i].ref++;
      g->whiteNodes--;
/*
      printf("Set %d color to %d:%d\n", i, BLACK, 
              g->adjList[i].color);
*/
      markAdjNodesGrey(g, g->nodeTable[i].nodeId, FALSE);
		}
  }

  for (i = 0; i <= g->maxNodeId; i++) {
    if (!isValidNode(i, &(g->nodeTable[i])))
      continue;

    if (g->nodeTable[i].color != WHITE)
      count++;
  }

/*
  printf("Nodes in bvec %d cover nodes %d\n", bvec->count, count);
*/
  return count;
}

#if 0
boolean isNodeDSRedundant(Graph *g, bitvec_t *bvec, int nodeId)
{
  int             i, degree = g->nodeTable[nodeId].degree;

/*
  printf("%s: Node ID = %d, Degree = %d\n", __FUNCTION__, nodeId, degree);
*/
  if (g->nodeTable[nodeId].ref < 2)
    return FALSE;

  for (i = 0; i < degree; i++) {
/*
    printf("Checking if nodeId %d is in bvec or neighbor of a node in bvec\n",
            nodeId);
*/
    if (g->nodeTable[g->neighborIds[nodeId][i]].ref < 2)
      return FALSE;
  }

/*
  printf("Node Id %d is %s\n", origNodeId, red ? "Redundant" : "Not Redundant");
*/

  return TRUE;
}
#endif /* 0 */

/* For capacitated dominating set, we need to find whether the node */
/* and all the nodes it actually covered are covered by other nodes */
boolean isNodeDSRedundant(Graph *g, bitvec_t *bvec, int nodeId)
{
  int             i, degree = g->nodeTable[nodeId].nodesCov.size, neighbor;

/*
  printf("%s: Node ID = %d, Degree = %d\n", __FUNCTION__, nodeId, degree);
  if (g->nodeTable[nodeId].ref < 2)
    return FALSE;
*/

  for (i = 0; i < degree; i++) {
    neighbor = g->nodeTable[nodeId].nodesCov.set[i];
    if (g->nodeTable[neighbor].ref < 2)
      return FALSE;
  }

  return TRUE;
}

/* Function to determine if a given set is a dominating set         */
/* We take each node in the given set, walk its adjacency list      */
/* adding these nodes to a different set. If at the end of walking  */
/* all the adjacency lists, we find that the number of nodes in the */
/* set is equal to the number of nodes in the graph, this is a      */
/* Dominating Set for the graph. This is done when the set to be    */
/* dominated is the entire graph. If the setTobeDominated is given, */
/* we check if the nodes walked are exactly the same as the given   */
/* set. If so, dominating is TRUE. Otherwise, we return FALSE.      */
boolean isSetDominating(Graph *g, nodeSet *S, nodeSet *toBeDominated)
{
  int             i;
  nodeSet         D;

  initNodeSet(&D);

  for (i = 0; i < S->size; i++) {
    addNodetoNodeSet(S->set[i], &D);
    g->nodeTable[S->set[i]].color = BLACK;
    g->nodeTable[S->set[i]].ref++;
    g->whiteNodes--;
/*
    printf("Set %d color to %d:%d\n", S->set[i], BLACK, 
            g->adjList[S->set[i]].color);
*/
    addAdjListNodestoSet(g, S->set[i], &D);
  }
/*
  printf("D.size = %d\n", D.size);
*/

	if (toBeDominated == NULL) {
    if (D.size == g->nodes)
      return TRUE;
    else
      return FALSE;
	} else {
    if (setSubset(&D, toBeDominated))
		  return TRUE;
		else
			return FALSE;
	}
}

/* Function to find and return the Maximal Independent Set of a graph */
void maximalIS(Graph *g, nodeSet *IS)
{
  int          i = 0;

  while (g->whiteNodes != 0) {
    if (!isValidNode(i, &(g->nodeTable[i]))) {
      i++;
      continue;
    }

    if (g->nodeTable[i].color == WHITE) {
      g->nodeTable[i].color = BLACK;
      g->whiteNodes--;
      addNodetoNodeSet(i, IS);
      addNodetoNodeSet(i, &(g->nodeTable[i].nodesCov));
      markAdjNodesGrey(g, g->nodeTable[i].nodeId, FALSE);
    }
    i++;
  }

  return;
}

/* Function to find and return a Maximal Independent Set of a graph */
/* given a starting index.                                          */
void MIS(Graph *g, nodeSet *IS, int i)
{
  while (g->whiteNodes != 0) {
    if (!isValidNode(i, &(g->nodeTable[i]))) {
      i = (i+1) % N;
      continue;
    } else if (g->nodeTable[i].color == WHITE) {
      g->nodeTable[i].color = BLACK;
      g->whiteNodes--;
      addNodetoNodeSet(g->nodeTable[i].nodeId, IS);
      markAdjNodesGrey(g, g->nodeTable[i].nodeId, FALSE);
    }
    i = (i+1) % N;
  }

  return;
}


#if 0
/* Function to determine if a set is an independent set            */
/* Needed for the exact MIDS as also the PTAS MIDS algorithms      */
/* Returns the pair which is adjacent in n1 and n2                 */
boolean isSetIndependent(Graph *g, nodeSet *S, int *n1, int *n2)
{
  int             i, j;

	if ((n1 != NULL) && (n2 != NULL)) {
  	*n1 = *n2 = -1;
	}

  for (i = 0; i < (S->size-1); i++) {
    for (j = (i+1); j < S->size; j++) {
      if (isNeighbor(g, S->set[i], S->set[j])) {
				if ((n1 != NULL) && (n2 != NULL)) {
  				*n1 = S->set[i];
  				*n2 = S->set[j];
				}
        return FALSE;
			}
    }
  }

  return TRUE;
}

/* Function that finds the conflicting nodes of the dominating set */
/* returns them in the parameter 'CS' which is used for repairing  */
/* independence between nodes of the dominating set                */
void constructConflictSet(Graph *g, nodeSet *DS, nodeSet *CS)
{
  int             i, j;

  for (i = 0; i < (DS->size-1); i++) {
    for (j = (i+1); j < DS->size; j++) {
      if (isNeighbor(g, DS->set[i], DS->set[j])) {
          addNodetoNodeSet(DS->set[i], CS);
          break;
				}
			}
    }

  return;
}

/* Function to construct the k-hop neighborhood of a given node 'v' */
/* in a given graph, G and returns it in a node set as output.      */
void kHopNeighbors(int v, Graph *G, int k, nodeSet *Neighbors)
{
  int       i, start = 0, end = 0;
  adjNode  *node;

  addNodetoNodeSet(v, Neighbors);
  /* If this node has no neighbors, i.e., it is disconnected, return */
  if (G->adjList[v].next == NULL)
    return;

  start = 0;
  end = Neighbors->size;
  while (k != 0) {
    for (i = start; i < end; i++) {
      /* Walk the adjacency list of the nodes from start to end */
      /* and add them to the neighborhood set 'Neighbors'.      */
      node = G->adjList[Neighbors->set[i]].next;
      while (node != NULL) {
        addNodetoNodeSet(node->nodeId, Neighbors);
        node = node->next;
      }
    }
    start = end;
    end = Neighbors->size;
    k--;
  }
}

/* Function to return the kth-hop neighbors of a given node 'v' */
void kthHopNeighbors(int v, Graph *G, int k, nodeSet *Neighbors)
{
  int       i, start = 0, end = 0, kMinus1Offset, kMinus1;
  adjNode  *node;
  nodeSet   kMinus1N;

  initNodeSet(&kMinus1N);
  initNodeSet(Neighbors);

  kMinus1 = k - 1;

  addNodetoNodeSet(v, &kMinus1N);
  /* If this node has no neighbors, i.e., it is disconnected, return */
  if (G->adjList[v].next == NULL)
    return;

  start = 0;
  end = kMinus1N.size;
  while (kMinus1 != 0) {
    for (i = start; i < end; i++) {
      /* Walk the adjacency list of the nodes from start to end */
      /* and add them to the neighborhood set 'Neighbors'.      */
      node = G->adjList[kMinus1N.set[i]].next;
      while (node != NULL) {
        addNodetoNodeSet(node->nodeId, &kMinus1N);
        node = node->next;
      }
    }
    start = end;
    end = kMinus1N.size;
    kMinus1--;
    if (kMinus1 == 1) 
      kMinus1Offset = end;
  }

  /* We have got all neighbors up to the (k-1)th hop. Now, we get the  */
  /* neighbors of the (k-1)th hop using the kMinus1Offset which will   */
  /* give us the kth hop neighbors. We need to make sure that these    */
  /* kth hop neighbors are not already present in the previous (k-1)th */
  /* set of neighbors. This is to ensure that the nodes selected are   */
  /* truly k-hops away from the given node 'v'.                        */
  start = kMinus1Offset;
  end = kMinus1N.size;
  
/*
  printf("kMinus1Offset = %d, end = %d\n", kMinus1Offset, end);
*/
  for (i = start; i < end; i++) {
    node = G->adjList[kMinus1N.set[i]].next;
    while (node != NULL) {
      if (!isSetMember(node->nodeId, &kMinus1N))
        addNodetoNodeSet(node->nodeId, Neighbors);
      node = node->next;
    }
  }

  return;
}

static int copyNode(Graph *newG, adjNode *old, adjNode *new)
{
  adjNode         *ptr = NULL;
  int              rc = 0;

  new->nodeId = old->nodeId;
  new->color = old->color;
  ptr = old->next;
  while (ptr != NULL) {
    rc = addAdjListNode(newG, old->nodeId, ptr->nodeId);
    if (rc == -1)
      return rc;
    ptr = ptr->next;
  }

  return rc;
}

int copyGraph(Graph *g, Graph *new)
{
  int            i, rc = 0;

  new->nodes     = g->nodes;
  new->maxNodeId = g->maxNodeId;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (isValidNode(i, &(g->adjList[i]))) {
      rc = copyNode(new, &(g->adjList[i]), &(new->adjList[i]));
      if (rc == -1)
        return rc;
    }
  }

  return rc;
}

static void delAdjList(adjNode *node)
{
  adjNode *cur = node->next, *next = NULL;

  while (cur != NULL) {
    next = cur->next;
    free(cur);
    cur = next;
  }

  return;
}

void delGraph(Graph *g)
{
  int            i;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (isValidNode(i, &(g->adjList[i])))
        delAdjList(&(g->adjList[i]));
  }

  return;
}

/* Function to return the nodeId of the first valid node in the graph */
int findFirstValidNode(Graph *G)
{
  int       i;

  if (G->nodes == 0) return -1;

  for (i = 0; i <= G->maxNodeId; i++) {
    if (isValidNode(i, &(G->adjList[i])))
      return(G->adjList[i].nodeId);
  }

  /* Should never come here */
  return -1;
}

/* Function to update the value of the maximum node ID in the graph */
/* if the node being deleted is the node with the maxNodeId         */
static void updateMaxNodeId(Graph *g)
{
  int      i;

  /* If there are no nodes in the graph, then invalidate maxNodeId */
  /* Else, if the no. of nodes is 1 and the first node is valid,   */
  /* that is the maxNodeId.                                        */
  if (g->nodes == 0) {
    g->maxNodeId = -1;
    return;
  } else if ((g->nodes == 1) && (g->adjList[0].nodeId == 0)) {
    g->maxNodeId = 0;
    return;
  }

  /* Go back from maxNodeId until you find a valid node ID.        */
  /* This will be the new maxNodeId.                               */
  for (i = g->maxNodeId; i > 0; i--) {
    if (g->adjList[i].nodeId != 0) {
      g->maxNodeId = i;
      return;
    }
  }

  /* Should never reach here */
  return;
}

/* When a node is deleted from the graph, it should also be removed   */
/* from the adjacency lists of its neighbors. This function does that */
/* 'nodeId' has to be removed from the adjacency list of 'adjNodeId'  */
static void delNodefromNeighborAdjList(Graph *g, int adjNodeId, int nodeId)
{
  adjNode      *neighAdjList = NULL, *ptr = NULL;

  neighAdjList = g->adjList[adjNodeId].next;
  while((neighAdjList != NULL) && (neighAdjList->nodeId != nodeId)) {
    ptr = neighAdjList;
    neighAdjList = neighAdjList->next;
  }

  /* Walked to the end of the list and could not find the node */
  if (neighAdjList == NULL) return;

  if (ptr == NULL) {   /* Node at the head of the list */
    g->adjList[adjNodeId].next = neighAdjList->next;
  } else {
    ptr->next = neighAdjList->next;
  }

  free(neighAdjList);

  /* Need to update the degree of the node here */
  g->adjList[adjNodeId].degree--;
  return;
}

/* Function to delete a node from the graph */
void delNodefromGraph(Graph *g, int nodeId)
{
  adjNode     *list, *temp;

  if (!isValidNode(nodeId, &(g->adjList[nodeId]))) {
    fprintf(stderr, "Node ID %d does not exist in the graph\n", nodeId);
    return;
  }

  /* Delete all entries in the adjacency list of this node ID   */
  /* after deleting this node's entry in the adjacency list of  */
  /* the neighbors.                                             */
  list = g->adjList[nodeId].next;
  while (list != NULL) {
    delNodefromNeighborAdjList(g, list->nodeId, nodeId);
    temp = list;
    list = list->next;
    free(temp);
  }

  /* Update the node ID value to reflect that this entry is not valid */
  if (nodeId == 0)
    g->adjList[nodeId].nodeId = -1;
  else
    g->adjList[nodeId].nodeId = 0;

  /* Update the number of nodes in the graph and the maxNodeId values */
  g->nodes--;
  if (nodeId == g->maxNodeId) {
    updateMaxNodeId(g);
  }

  return;
}

#endif /* 0 */

