#ifndef __GRAPH_H
#define __GRAPH_H

/* The node in the adjacency matrix */
typedef struct Node_s {
  int               nodeId;     /* ID of the node                           */
  short             degree;     /* The degree of a node                     */
  int               capacity;   /* Capacity of a node to cover N(v)         */
  int               resCap;     /* Residual capacity of a node              */
  float             weight;     /* Pheromone of a node                      */
  float             pFact;      /* Pheromone factor of a node (inc. ETA)    */
  int               ref;        /* Maintains a count of how many nodes are  */
                                /* dominating it, including itself.         */
  colour_t          color;      /* Used to determine if a node is dominated */
  nodeSet           nodesCov;   /* Nodes covered by this node               */
} Node_t;


/* Graph as an adjacency Matrix */
/* The node ID is used as an index into the Matrix.                    */
/* However, since it is possible that the number of nodes is less than */
/* the maximum node ID, we maintain a different field for maxNodeId.   */
typedef struct graph_s {
  int         nodes;
  int         maxNodeId;
  Node_t      nodeTable[N];
  char        adjMatrix[N][N];
  int         neighborIds[N][N];
  int         whiteNodes;
} Graph;

/************** APIs to manipulate the graph **************/

/* Initialize the graph structure                                         */
void initGraph(Graph *g);

/* Initialize the color of every node in the graph to WHITE               */
void initColorWhite(Graph *g);

/* Calculate the number of white nodes adjacent to the given node         */
/* limited by the capacity of the node.                                   */
int calcWhiteDegCovbyCap(Graph *g, Node_t *node);

/* Calculate the number of white nodes adjacent to the given node         */
int calcWhiteDegree(Graph *g, Node_t *node);

void printNodeColors(Graph *g);
void printWhiteNodes(Graph *g);

/* Mark the adjacency list of the given node GREY                         */
/* If cap == TRUE, select the first 'capacity' nodes of the adjacency     */
/* list as the nodes covered by the given node.                           */
void markAdjNodesGrey(Graph *g, int nodeId, boolean cap);

/* Mark the adjacency list of the given node GREY selecting neighbors     */
/* in a random fashion until the residual capacity of the node is zero.   */
void markCapAdjNodesGrey(Graph *g, int nodeId, boolean useAllCap);

/* Mark the adjacent nodes up to 'capacity' GREY where the adjacent nodes */
/* are sorted in increasing order of degree.                              */
void markLdegAdjNodesGrey(Graph *g, int nodeId);

/* Find a maximal Independent Set in the given graph                      */
void maximalIS(Graph *g, nodeSet *IS);

/* Find the maximal independent set starting from node with index 'i'     */
void MIS(Graph *g, nodeSet *IS, int i);

#if 0
/* Mark 'capacity' adjacent nodes of nodeId GREY                          */
void markCapAdjNodesGrey(Graph *g, int nodeId);
#endif

/* Check if a given index represents a valid node in the graph            */
boolean isValidNode(int index, Node_t *node);

/* Function to add a node to a weighted graph                             */
void addWtNodetoGraph(Graph *g, int nodeId, float weight);

/* Function to add a node to a Capacitated graph                          */
void addCapNodetoGraph(Graph *g, int nodeId, int capacity);

/* Add an edge to the graph. This might involve adding the from and to    */
/* nodes to the graph, if they are not already in the graph. It also      */
/* involves updating the adjacency lists of the nodes.                    */
int addEdgetoGraph(Graph *g, int fromNode, int toNode);

/* Add a directed edge to the graph. This means that only the edge from   */
/* 'fromNode' to 'toNode' is added to the graph and not vice versa.       */
int addDirEdgetoGraph(Graph *g, int fromNode, int toNode);

/* Add a directed edge from 'fromNode' to 'toNode' in a graph with        */
/* uniform  capacities for the nodes.                                     */
int addDirEdgetoCapGraph(Graph *g, int fromNode, int toNode, int capacity);

/* Function to construct the set of nodes 'V' of a given graph            */
void constructSet(Graph *g, nodeSet *V);

/* Function to determine if n2 is a neighbor of n1                        */
boolean isNeighbor(Graph *G, int n1, int n2);

#if 0
/* Function to copy one graph to the other                                */
int copyGraph(Graph *g, Graph *new);

/* Function to delete the graph g                                         */
void delGraph(Graph *g);

/* Function to delete a given node from the graph                         */
void delNodefromGraph(Graph *g, int nodeId);

/* Function that returns the first valid node ID of the given graph       */
int findFirstValidNode(Graph *G);

/* Function to find the set of nodes that are not independent in DS       */
/* and return them in the conflict set 'CS'.                              */
void constructConflictSet(Graph *g, nodeSet *DS, nodeSet *CS);

/* Function to calculate the maximal Independent Set using a greedy algo. */
void maximalIS(Graph *g, nodeSet *IS);

/* Function to return MIS given a starting nodeId                         */
void MIS(Graph *g, nodeSet *IS, int nodeId);

/* Function to remove the closed neighborhood of a node 'v'               */
void removeClosedNofV(Graph *g, int v);

/* Function to remove closed neighborhood of a set from the graph 'g'     */
void removeClosedNofSet(Graph *g, nodeSet *D);
#endif

/* Given a bit vector, this function returns the number of nodes covered  */
/* by that bit vector.                                                    */
int nodesCoveredbyVec(Graph *g, bitvec_t *bvec);

/* Get the set of nodes in the graph which are coloured WHITE             */
void getWhiteNodeSet(Graph *g, nodeSet *S);

/* Function to determine if a given set is dominating or not              */
/* If 'toBeDominated' is NULL, we need to verify if 'S' dominated 'g'     */
/* Otherwise, verify that 'S' dominates 'toBeDominated'.                  */
boolean isSetDominating(Graph *g, nodeSet *S, nodeSet *toBeDominated);

/* Function that determines if a given node is a redundant member of the  */
/* given dominating set in bvec - i.e., if the given node and all its     */
/* neighbors are neighbors of the other nodes of DS, it is not needed     */
/* in the dominating set.                                                 */
boolean isNodeDSRedundant(Graph *g, bitvec_t *D, int nodeId);

/* Function to print the adjacency lists of the nodes (for debug only)    */
void printGraph(Graph *g);

#endif /* __GRAPH_H */
