#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>

#include "gen.h"
#include "bitv.h"
#include "popu_t.h"
#include "set.h"
#include "graph_am.h"

#define MAX_TRIALS             50
#define MAX_NO_CHANGE_TRIALS   50000

#define PROB_INIT              0.5
#define PROB_CROSSOVER         0.9
#define PROB_MUTATION          0.002
#define PROB_BETTER            0.8
#define PROB_HEURISTIC         0.6
#define PROB_FILTER            0.4

#define SEED_PERCENT           0.2

/******* GLOBAL VARIABLES  *******/
static Graph           g;
static population_t    pop;
static int             gen = 0;

#if 0
static int             avgDSize = 0, redNodes = 0, mutate = 0;
#endif /* 0 */


/************ Function prototypes **********/
extern int parse_udg_cap(FILE *fp, Graph *g, int var);

static int genInitialPopulation(int *totalMembers, int vertices);
static int crossover(int initMembers, int vertices, bitvec_t *bvec,
                     int *fitness);
static void mutate(Graph *g, bitvec_t *bvec, int *fitness);
static int genNewChild(int initMembers);
static int heurCapMDS(Graph *g, bitvec_t *bvec, int *fitness);
static int randomCapMDS(Graph *g, bitvec_t *bvec, int *fitness);
static void minimizeDS(bitvec_t *bvec, int *fitness); 
static void addNodetoDS(bitvec_t *bvec, int *fitness, int nodeId);
static void delNodefromDS(bitvec_t *bvec, int *fitness, int nodeId);
static void printNodesCov(Graph *g, bitvec_t *bvec);

int main(int argc, char **argv)
{
  FILE           *fp = NULL;
  int             initialPopulation = 0, bestGen = -1;
  int             vertices = 0, f, noChangeTrials = 0;
  int             rc, maxGen, dup = 0, trials = 0;
/*  int             gen = 0; */
  float           bestFitness;
  popu_member_t  *member = NULL;
  unsigned long   numSecs;
  struct timeval  t1, t2;

  if (argc < 5) {
    fprintf(stderr, 
            "Usage: ga-mds <filename> (U)ni-cap/(V)ar-cap <initial population> <maxGen>\n");
    exit(1);
  }

  if ((fp = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "Error opening file %s\n", argv[1]);
    exit(2);
  }

  initGraph(&g);
  if (argv[2][0] == 'u' || argv[2][0] == 'U')
    rc = parse_udg_cap(fp, &g, 0);
  else if (argv[2][0] == 'v' || argv[2][0] == 'V')
    rc = parse_udg_cap(fp, &g, 1);
  else {
    fprintf(stderr, "Error in file format parameter\n");
    exit(3);
  }
 
  if (rc == -1) {
    fprintf(stderr, "Error parsing the input file\n");
    exit(4);
  }
  vertices = g.nodes;

/*
  printGraph(&g);
*/
  
  initialPopulation  = atoi(argv[3]);
  maxGen             = atoi(argv[4]);

  gettimeofday(&t1, NULL);

  popuInit(&pop);
  rc = genInitialPopulation(&initialPopulation, vertices);
  if (rc < 0) {
    fprintf(stderr, "Error generating initial population\n");
    exit(5);
  }
  bestFitness = popuBestMember(&pop, &member);
/*
  printf("After initial population, best member fitness = %f\n", bestFitness);
*/

  while (gen < maxGen) {
    if ((trials == MAX_TRIALS) || (noChangeTrials == MAX_NO_CHANGE_TRIALS))
      break;

/*
    printf("Gen %d: ", gen);
*/
    rc = genNewChild(initialPopulation);
    if (rc < 0) {
      dup++;
      trials++;
    } else {
      gen++;
      trials = 0;

      if ((f = popuBestMember(&pop, &member)) < bestFitness) {
        bestGen = gen;
        bestFitness = f;
        noChangeTrials = 0;
      } else 
        noChangeTrials++;
/*
      printf("Gen %d: Fitness = %d\n", gen, f);
*/
    }
  }

/*
  if (gen != 0)
    avgDSize = avgDSize / gen;
  printf("Avg.DS = %d, mutate = %d, redNodes = %d, DSRed = %ld\n", 
          avgDSize, mutate, redNodes, dsRedSecs);

  printf("Best Gen = %d, Current gen = %d\n", bestGen, gen);
  printf("Real best fitness = %d\n", realBestFitness);
  popuPrint();
  printf("GenNewChild completed over %d generations with dup = %d\n", gen, dup);
*/

  gettimeofday(&t2, NULL);
  numSecs = t2.tv_sec - t1.tv_sec;
  printf("%s %d %ld\n", argv[1], (int) popuBestMember(&pop, &member), numSecs);

/*
  printf("Members of the solution are \n");
  bitPrint(&(member->vec));
*/

  exit(0);
}


/* Function to generate initial population with the given number of    */
/* members. 'vertices' gives the length of the bitvector to generate.  */
static int genInitialPopulation(int *totalMembers, int vertices)
{
  int          i = 0, j, rc, heurMembers;
  bitvec_t     bvec;
  double       myrand;
  int          fitness = 0;
  int          dup = 0, trials = 0;

/*
  time_t       curtime;
  int          index;

  srand(time(&curtime)); 
  trials = 0;
*/

  /* Generate PROB_SEED members using heuristic for CAPMDS so that   */
  /* there are enough members with smaller cardinality that help     */
  /* during choice of better parents. This can lead to faster and    */
  /* better convergence.                                             */
  /* For better variety, create the remaining members randomly and   */
  /* repair the dominating property if necessary.                    */
  heurMembers = (int) (SEED_PERCENT * (*totalMembers));
  while (i < heurMembers) {
    fitness = 0;
    bitInit(&bvec, vertices);
    initColorWhite(&g);
    heurCapMDS(&g, &bvec, &fitness);
    rc = popuAdd(&pop, &bvec, fitness);
    i++;
/*
    printf("Member %d: f = %d\n", i, fitness);
*/
  }

  while (i < *totalMembers) {
    if (trials == MAX_TRIALS) {
      *totalMembers = i;
      fprintf(stderr, "Total members modified to %d\n", *totalMembers);
      break;
    }

    fitness = 0;
    bitInit(&bvec, vertices);
    initColorWhite(&g);
/*
    printf("member = %d, G.nodes = %d, G.maxNodeId = %d, G.whiteNodes = %d\n",
           i, g.nodes, g.maxNodeId, g.whiteNodes);
*/

    /* Set the bits in the bit vector randomly */
    for (j = 0; j < vertices; j++) {
      myrand = rand() / (RAND_MAX + 1.0);
      if (myrand < PROB_INIT) {
        addNodetoDS(&bvec, &fitness, j);
      }
    }

/*
    printf("Bit vector count = %d, Fitness = %d\n", bvec.count, fitness);
    printf("Init member %d: White nodes = %d\n", i, g.whiteNodes);
*/

    /* Check if the member generated is dominating; repair it otherwise */
    /* Check if it is unique; add it to the population if it is         */
    if (g.whiteNodes != 0) {
/*
      printf("%s: Population member %d is not dominating\n", __FUNCTION__, i);
      printNodeColors(&g);
*/

      myrand = rand() / (RAND_MAX + 1.0);
      if (myrand <= PROB_HEURISTIC) {
/*
        printf("Calling heurCapMDS....\n"); 
*/
        heurCapMDS(&g, &bvec, &fitness);
      } else {
/*
        printf("Calling randomCapMDS for init member = %d, whiteNodes = %d\n", 
                i, g.whiteNodes); 
*/
        randomCapMDS(&g, &bvec, &fitness);
      }

/*
      printf("%s: Used heur/random to make IS dominating...minimizing it\n",
             __FUNCTION__);
      bitPrint(&bvec);
*/
    }

/*
    printf("Init member %d: bvec count = %d:%d, White nodes after dom = %d\n", 
            i, bvec.count, fitness, g.whiteNodes);
*/

    /* Minimize the DS before adding it to the population */
    minimizeDS(&bvec, &fitness);

/*
    printf("%s: minimized bit vector = \n", __FUNCTION__);
    bitPrint(&bvec);
*/
    if (popuExistMember(&pop, &bvec, fitness)) {
/*
      printf("Duplicate member found....\n");
*/
      dup++;
      trials++;
      continue; 
    }

    if ((rc = popuAdd(&pop, &bvec, fitness)) < 0) {
      fprintf(stderr, "%s: Error returned from popuAdd\n", __FUNCTION__);
      return rc;
    } 
/*
    else
    printf("Added member %d with fitness %d to the population\n", i, fitness);
*/
    trials = 0;
    i++;
  }

/*
  printf("No. of dups during initial population of %d members = %d\n", i, dup);
*/

  return 0;
}

static int genNewChild(int initMembers)
{
  int               rc = 0, index, vertices;
  bitvec_t          bvec;
  float             myrand; 
  int               fitness = 0;

  vertices = g.nodes;
  initColorWhite(&g);

  rc = 0;
  bitInit(&bvec, vertices);

  myrand = ((float) rand()) / ((float) (RAND_MAX + 1.0));
  if (myrand <= PROB_CROSSOVER) {
    crossover(initMembers, vertices, &bvec, &fitness);
    mutate(&g, &bvec, &fitness);
  }
  else {
    /* Generate a child randomly */
    for (index = 0; index < vertices; index++) {
      myrand = rand() / (RAND_MAX + 1.0);
      if (myrand < PROB_INIT) {
        addNodetoDS(&bvec, &fitness, index);
      }
    }
  }

  /* If the child so generated is not dominating, repair it */
  /* Check for uniqueness, delete the tail and then add it  */
  /* to the existing population                             */
  if (g.whiteNodes != 0) {
    myrand = rand() / (RAND_MAX + 1.0);
    if (myrand <= PROB_HEURISTIC) {
      heurCapMDS(&g, &bvec, &fitness);
    } else {
      randomCapMDS(&g, &bvec, &fitness);
    }

/*
    printf("After repair, bitvec is given below\n");
    bitPrint(&bvec);
*/
  }


  minimizeDS(&bvec, &fitness);

/*
  printf("After minimize bitvec is \n");
  bitPrint(&bvec);
  if (popuExistMember(&pop, &bvec, fitness))
  {
    fprintf(stderr, "%s: Crossover child is not unique\n", __FUNCTION__);
    rc = -1;
  }
  printf("%d\n", fitness);
*/

  rc = popuReplace(&pop, &bvec, fitness);

  return rc;
}

static int crossover(int initMembers, int vertices, bitvec_t *bvec,
                     int *fitness)
{
  double             probBetterParent;
  float              myrand;
  int                i, index;
	popu_member_t     *p1 = NULL, *p2 = NULL;
  popu_member_t     *parent1 = NULL, *parent2 = NULL;

  index = rand() % initMembers;
  p1 = &(pop.table[index]);

  index = rand() % initMembers;
  p2 = &(pop.table[index]);

  myrand = ((float) rand()) / ((float) (RAND_MAX + 1.0));
  if (myrand < PROB_BETTER) {
    if (p1->fitness < p2->fitness)
      parent1 = p1;
    else
      parent1 = p2;
  } else {
    if (p1->fitness < p2->fitness)
      parent1 = p2;
    else
      parent1 = p1;
  }

/*
  printf("P1.fitness = %d, P2.fitness = %d, parent1.fitness = %d\n",
          p1->fitness, p2->fitness, parent1->fitness);
*/

  index = rand() % initMembers;
  p1 = &(pop.table[index]);

  index = rand() % initMembers;
  p2 = &(pop.table[index]);

  myrand = ((float) rand()) / ((float) (RAND_MAX + 1.0));
  if (myrand < PROB_BETTER) {
    if (p1->fitness < p2->fitness)
      parent2 = p1;
    else
      parent2 = p2;
  } else {
    if (p1->fitness < p2->fitness)
      parent2 = p2;
    else
      parent2 = p1;
  }

/*
  printf("P1.fitness = %d, P2.fitness = %d, parent2.fitness = %d\n",
          p1->fitness, p2->fitness, parent2->fitness);
*/

  if (parent1->fitness < parent2->fitness)
    probBetterParent = parent2->fitness / (parent1->fitness + parent2->fitness);
  else
    probBetterParent = parent1->fitness / (parent1->fitness + parent2->fitness);

  for (i = 0; i < vertices; i++) {
    myrand = rand() / (RAND_MAX + 1.0);
    if ((myrand < probBetterParent) && (bitIsSet(&(parent1->vec), i))) {
      addNodetoDS(bvec, fitness, i);
    } else if (bitIsSet(&(parent2->vec), i)) {
      addNodetoDS(bvec, fitness, i);
    }
    i++;
  }

  return 0;
}

static void mutate(Graph *g, bitvec_t *bvec, int *fitness)
{
  int             index, vertices = g->nodes;
  float           myrand;

  for (index = 0; index < vertices; index++) {
    myrand = ((float) rand()) / ((float) (RAND_MAX + 1.0));
    if (myrand <= PROB_MUTATION) {
      if (bitIsSet(bvec, index)) {
/*
        printf("Clearing node %d in mutation with nodes covered\n", index);
        printNodesCov(g, bvec);
*/
        delNodefromDS(bvec, fitness, index);
      } else {
        addNodetoDS(bvec, fitness, index);
      }
    }
  }

  return;
}

static void addNodetoDS(bitvec_t *bvec, int *fitness, int nodeId)
{
  boolean    red = TRUE;
  int        i;

/*
  printf("%s: Node ID = %d\n", __FUNCTION__, nodeId);
*/

  /* If a node and all its neighbors have a non-WHITE color, then */
  /* adding this node to the dominating set is redundant.         */
  if (g.nodeTable[nodeId].color != WHITE) {
    for (i = 0; i < g.nodeTable[nodeId].degree; i++) {
      if (g.nodeTable[g.neighborIds[nodeId][i]].color == WHITE) {
        red = FALSE;
        break;
      }
    }
  } else
    red = FALSE;

  if (red == TRUE) {
/*
    printf("%d is already redundant...so not adding it to DS\n", nodeId);
    redNodes++;
*/
    return;
  }

  /* Add the node to the bit vector, set its color to BLACK, all */
  /* its neighbors to GREY and update white Nodes                */
  bitSet(bvec, nodeId);
  if (g.nodeTable[nodeId].color == WHITE) {
    g.whiteNodes--;
/*
    g.nodeTable[nodeId].resCap--;
    addNodetoNodeSet(nodeId, &(g.nodeTable[nodeId].nodesCov));
    g.nodeTable[nodeId].ref++;
*/
  }
  addNodetoNodeSet(nodeId, &(g.nodeTable[nodeId].nodesCov));
  g.nodeTable[nodeId].ref++;
  g.nodeTable[nodeId].color = BLACK;
  markCapAdjNodesGrey(&g, nodeId, TRUE);
  (*fitness)++;
/*
  printf("Added %d to DS, white Nodes = %d\n", nodeId, g.whiteNodes);
*/

  return;
}

static void delNodefromDS(bitvec_t *bvec, int *fitness, int nodeId)
{
  int            i, nid;

  bitClear(bvec, nodeId);

#if 0
  g.nodeTable[nodeId].ref--;

  /* If bit is being cleared as part of mutation, it may not be   */
  /* a redundant node. Hence, when we decrement ref, it may be 0  */
  /* In such a case, the color should be changed to WHITE. If     */
  /* bit is being cleared from minimizeDS, it is a redundant node */
  /* So, the color should be changed to GREY as some other node   */
  /* will be referring to it.                                     */
  if (g.nodeTable[nodeId].ref == 0) {
    g.nodeTable[nodeId].color = WHITE;
    g.whiteNodes++;
  } else {
    g.nodeTable[nodeId].color = GREY;
  }
#endif /* 0 */

  /* Now, update the reference of the adj. nodes and change their */
  /* color accordingly. As earlier, for mutation the color may    */
  /* become WHITE whereas in minimization, they will remain GREY  */
  for (i = 0; i < g.nodeTable[nodeId].nodesCov.size; i++) {
    nid = g.nodeTable[nodeId].nodesCov.set[i];
    g.nodeTable[nid].ref--;
    if (g.nodeTable[nid].ref == 0) {
      g.nodeTable[nid].color = WHITE;
      g.whiteNodes++;
    } else {
      g.nodeTable[nid].color = GREY;
    }
  }

  /* Re-initialize the nodesCov set for this node as it no longer */
  /* covers any nodes. This is important esp. in mutation because */
  /* : let us say that it is cleared and so some nodes become     */
  /* WHITE; then, it is selected again as part of repairing the   */
  /* DS; now, it will add nodes to the nodesCov to the existing   */
  /* set which is wrong. Hence, we need to clear the nodesCov     */
  /* We also need to reset the resCap.                            */
  g.nodeTable[nodeId].resCap = g.nodeTable[nodeId].capacity;
  initNodeSet(&(g.nodeTable[nodeId].nodesCov));

  /* Update the fitness value and return                          */
  (*fitness)--;

  return;
}

static int heurCapMDS(Graph *g, bitvec_t *bvec, int *fitness)
{
  int           i, maxCap, cap, maxNode = -1;
  Node_t       *node;

/*
  printf("%s: Set to be made dominating\n", __FUNCTION__);
  bitPrint(bvec);
*/

  while (TRUE) {
    maxCap = 0;
    cap = 0;

    for (i = 0; i <= g->maxNodeId; i++) {
      if (!isValidNode(i, &(g->nodeTable[i])))
        continue;

      node = &(g->nodeTable[i]); 
      if (node->color != BLACK) {
        cap = calcWhiteDegCovbyCap(g, node);
        if (maxCap < cap) {
          maxCap = cap;
          maxNode = node->nodeId;
/*
          printf("%d: Node with max white degree (%d) is %d\n", i,
                  cap, maxNode);
*/
        } else if (maxCap == cap) {
          if (g->nodeTable[i].degree < g->nodeTable[maxNode].degree) {
            maxCap = cap;
            maxNode = node->nodeId;
          }
        }
      }
    }
    /* If no node with white degree is found, then, no node is white */
    if (maxCap == 0) {
/*
      printf("No node with white degree is found\n");
*/
      break;
    }

    /* Mark all neighbors of node 'maxNode' GREY */
    bitSet(bvec, maxNode);
    if (g->nodeTable[maxNode].color == WHITE) {
      g->whiteNodes--;
/*
      g->nodeTable[maxNode].resCap--;
      addNodetoNodeSet(maxNode, &(g->nodeTable[maxNode].nodesCov));
      g->nodeTable[maxNode].ref++;
*/
    }
    addNodetoNodeSet(maxNode, &(g->nodeTable[maxNode].nodesCov));
    g->nodeTable[maxNode].ref++;
    g->nodeTable[maxNode].color = BLACK;
    markCapAdjNodesGrey(g, maxNode, TRUE);
    (*fitness)++;
  }

  return 0;
}

static int findWhiteNode(Graph *g)
{
  int      i;

  for (i = 0; i <= g->maxNodeId; i++) {
    if (!isValidNode(i, &(g->nodeTable[i])))
      continue;
    if (g->nodeTable[i].color == WHITE)
      return i;
  }

  return -1;
}

static int randomCapMDS(Graph *g, bitvec_t *bvec, int *fitness)
{
  int            index, nid, i;
  nodeSet        S;
  boolean        red;

  initNodeSet(&S);
  for (index = 0; index < bvec->members; index++) {
    red = TRUE;
    if (bvec->vec[index] == 0) {
      if (g->nodeTable[index].color != WHITE) {
        for (i = 0; i < g->nodeTable[index].degree; i++) {
          nid = g->neighborIds[index][i];
          if (g->nodeTable[nid].color == WHITE) {
            red = FALSE;
            break;
          }
        }
      } else {
        red = FALSE;
      }
      if (!red)
        addNodetoNodeSet(index, &S);
    }
  }

/*
  printf("Number of nodes not set in bit vector = %d\n", S.size);
  initColorWhite(g);
*/

  while (g->whiteNodes != 0) {

/*
    printf("Number of nodes not in bvec with count %d = %d\n",
            bvec->count, S.size);
    printNodeSet(&S);
    printf("White Nodes = %d\n", g->whiteNodes);
*/

    if (S.size == 0) {
      nid = findWhiteNode(g);
      printf("Unexpected S.size = 0...node %d still has WHITE color\n", nid);
      if (nid != -1)
        printf("Bitvector value = %d\n", bvec->vec[nid]);
      exit(1);
    }

    index = rand() % S.size;
    nid = S.set[index];
    bitSet(bvec, nid);
/*
    printf("%d:Setting bit no. %d with color %d\n", index, S.set[index],
            g->nodeTable[nid].color);
*/

    if (g->nodeTable[nid].color == WHITE) {
      g->whiteNodes--;
/*
      g->nodeTable[nid].resCap--;
      addNodetoNodeSet(nid, &(g->nodeTable[nid].nodesCov));
      g->nodeTable[nid].ref++;
*/
    }
    addNodetoNodeSet(nid, &(g->nodeTable[nid].nodesCov));
    g->nodeTable[nid].ref++;
    g->nodeTable[nid].color = BLACK;
    markCapAdjNodesGrey(g, nid, TRUE);
    (*fitness)++;
    delNodefromNodeSet(nid, &S);
  }

/*
  printf("Number of bits set in bvec = %d\n", bvec->count);
*/
  return 0;
}

static void minimizeDS(bitvec_t *bvec, int *fitness)
{
  int               i, index; 
  int               maxCap, cap, capDeg, minDeg, deg;
  int               minId, nid;
  nodeSet           redSet;
  float             myrand;

/*
  printf("In Function %s, set input\n", __FUNCTION__);
  bitPrint(bvec);
  printNodesCov(&g, bvec);
*/

  /* Initialize and collect the set of nodes which are redundant   */
  /* in the given dominating set.                                  */
  while (TRUE) {
    initNodeSet(&redSet);
    for (i = 0; i < bvec->members; i++) {
      if (bvec->vec[i] == 0)
        continue;
      if (isNodeDSRedundant(&g, bvec, i)) {
/*
        printf("adding node %d to red set\n", i);
*/
        addNodetoNodeSet(i, &redSet);
      }
    }

    if (redSet.size == 0) {
      break;
    }

/*
    printf("%s: Redundant set of size %d is found\n", __FUNCTION__, 
           redSet.size);
    printNodeSet(&redSet);
*/

    myrand = ((float) rand()) / ((float) (RAND_MAX + 1.0));
    if (myrand <= PROB_FILTER) {
      index = rand() % redSet.size;
/*
      printf("Clearing redundant node %d from the bit vector\n", 
              redSet.set[index]);
*/
      delNodefromDS(bvec, fitness, redSet.set[index]);
    } else {
      minDeg = N;
      capDeg = N;
      maxCap = -1;
      nid = -1;
      minId = -1;
      for (i = 0; i < redSet.size; i++) {
        cap = g.nodeTable[redSet.set[i]].resCap;
        deg = g.nodeTable[redSet.set[i]].degree;
        if (cap > 0) {
          if (maxCap < cap) {
            maxCap = cap;
            capDeg = deg;
            nid = redSet.set[i];
          } else if ((maxCap == cap) && (deg < capDeg)) {
            capDeg = deg;
            nid = redSet.set[i];
          }
        } else if (minDeg > deg) {
          minDeg = deg;
          minId = redSet.set[i];
        }
      }

      if (nid != -1) {
/*
        printf("Node %d with resCap = %d found in redSet, removing it\n",
                nid, maxCap);
*/
        delNodefromDS(bvec, fitness, nid);
      } else {
/*
        printf("No node with resCap, removing %d with degree %d from redSet\n",
                minId, minDeg);
*/
        delNodefromDS(bvec, fitness, minId);
      }
    }
  }

  return;
}

static void printNodesCov(Graph *g, bitvec_t *bvec)
{
  int         i, j;

  for (i = 0; i < bvec->members; i++) {
    if (bvec->vec[i] == 0)
      continue;

    printf("Node %d covers: ", i);
    for (j = 0; j < g->nodeTable[i].nodesCov.size; j++) 
      printf("%d ", g->nodeTable[i].nodesCov.set[j]);
    printf("\n");
  }

  return;
}

