#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gen.h"
#include "bitv.h"
#include "set.h"
#include "graph_am.h"

/*************************************************************************/
/**************       PARSE UDG FILE FORMAT FUNCTION     *****************/
/*************************************************************************/

#define MAXBUF 10000

int parse_UDG(FILE *fp, Graph *g)
{
  char      str[MAXBUF];
  int       fromNode, toNode;

  while (fgets(str, MAXBUF, fp) != NULL) {
      sscanf(str, "%d %d", &fromNode, &toNode);
      addDirEdgetoGraph(g, fromNode, toNode);
  }

  return 0;
}


/*************************************************************************/
/**************  PARSE UDG CAPACITY FILE FORMAT FUNCTION *****************/
/*************************************************************************/
int parse_udg_cap(FILE *fp, Graph *g, int var)
{
  char      str[MAXBUF];
  int       fromNode, toNode, nodes;
  int       capacity, i;

  /* Number of nodes in the graph  */
  if (fgets(str, MAXBUF, fp) == NULL) {
    fprintf(stderr, "%s: Error parsing file\n", __FUNCTION__);
    return -1;
  }
  if (var == 0) {
    sscanf(str, "%d %d", &nodes, &capacity);
    /* Add the edges in the graph                       */
    while (fgets(str, MAXBUF, fp) != NULL) {
      sscanf(str, "%d %d", &fromNode, &toNode);
      addDirEdgetoCapGraph(g, fromNode, toNode, capacity);
    }
  } else {
    sscanf(str, "%d", &nodes);
    /* Parse the variable capacity of the nodes in the graph and    */
    /* add the node along with its weight to the graph              */
    for (i = 0; i < nodes; i++) {
      if (fgets(str, MAXBUF, fp) == NULL) {
        fprintf(stderr, "%s: Error parsing file\n", __FUNCTION__);
        return -1;
      }
      sscanf(str, "%d %d", &toNode, &capacity);
      addCapNodetoGraph(g, toNode, capacity);
    }

    /* Skip the blank line between capacities and edges */
    fgets(str, MAXBUF, fp);

    /* Add the edges in the graph                       */
    while (fgets(str, MAXBUF, fp) != NULL) {
      sscanf(str, "%d %d", &fromNode, &toNode);
      addDirEdgetoGraph(g, fromNode, toNode);
    }
  }

  return 0;
}


/*************************************************************************/
/**************     PARSE BRITE FILE FORMAT FUNCTION     *****************/
/*************************************************************************/
#define DELIM "()"

int parse_brite(FILE *fp, Graph *g)
{
  char      str[MAXBUF], tmp[32];
  char      *token = NULL;
  int       nodes, edges, fromNode, toNode, junk, i;

  if (fgets(str, MAXBUF, fp) == NULL) {
    fprintf(stderr, "Error reading the file\n");
    return -1;
  }

  token = strtok(str, DELIM);
  if (token == NULL) {
    fprintf(stderr, "Error parsing the file\n");
    return -1;
  }

  token = strtok(NULL, DELIM);
  if (token == NULL) {
    fprintf(stderr, "Error parsing the file\n");
    return -1;
  }

/*  printf("Token = [%s]\n", token); */
  sscanf(token, "%d %s %d %s", &nodes, tmp, &edges, tmp);
/*  printf("Nodes = %d, Edges = %d\n", nodes, edges); */

  while (fgets(str, MAXBUF, fp) != NULL) {
    if (str[0] != 'E')
      continue;

    for (i = 0; i < edges; i++) {
      fgets(str, MAXBUF, fp);
      sscanf(str, "%d %d %d", &junk, &fromNode, &toNode);
      addEdgetoGraph(g, fromNode, toNode);
    }
  }

  return 0;
}

/*************************************************************************/
/**************     PARSE DIMACS FILE FORMAT FUNCTION    *****************/
/*************************************************************************/
#define INT 32		/* computer word size */
#define CHARBITS 8
#define edge(x,y) (bitmap[y/CHARBITS][x] & (1<<(y%CHARBITS)))

#define MAX_NR_VERTICES		2400	/* = NMAX */
#define MAX_NR_VERTICESdiv8	300	/* = NMAX/8 */
#define BOOL	char
#define MAX_PREAMBLE 10000
#define REORDER

unsigned mask[INT] =
   {
   1, 1<<31, 1<<30, 1<<29, 1<<28, 1<<27, 1<<26, 1<<25, 1<<24,
   1<<23, 1<<22, 1<<21, 1<<20, 1<<19, 1<<18, 1<<17, 1<<16,
   1<<15, 1<<14, 1<<13, 1<<12, 1<<11, 1<<10, 1<<9, 1<<8,
   1<<7, 1<<6, 1<<5, 1<<4, 1<<3, 1<<2, 1<<1
   };  /* CAUTION - assumes 32 bit machine */

  /* graph input parameters */

int Nr_vert, Nr_edges;
BOOL Bitmap[MAX_NR_VERTICES][MAX_NR_VERTICESdiv8];
static char Preamble[MAX_PREAMBLE];
char masks[ 8 ] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };

unsigned char bitmap[N/CHARBITS+1][N+1];

static int get_params(void);
static int readgraph(Graph *g, FILE *fp);

int parse_dimacs(FILE *inputfile, Graph *g)
{
  int i,j, nodes;

  readgraph(g, inputfile);
  nodes = g->nodes;

  g->nodes = 0;
  for (i = 1; i <= nodes; i++) {
    for (j = 1; j <= nodes; j++) {
      if (edge(i, j))
        addEdgetoGraph(g, i, j);
    }
  }

	return 0;
}

BOOL get_edge( i, j )
int i,j;
{
	int byte, bit;
	char mask;
	int k;

	if (i<j) {
		k = i;
		i = j;
		j = k;
	}
	
	bit  = 7-(j & 0x00000007);
	byte = j >> 3;
	
	mask = masks[bit];
	return( (Bitmap[i][byte] & mask)==mask );
}

static int readgraph(Graph *g, FILE *fp) 
{
   int i,j;
   int length = 0;

	if (!fscanf(fp, "%d\n", &length)) { 
		fprintf(stderr, "ERROR: Corrupted preamble.\n"); 
		return -1;
	}

	if(length >= MAX_PREAMBLE) { 
		printf("ERROR: Too long preamble.\n"); 
		return -1;
	}
		   
	fread(Preamble, 1, length, fp);
	Preamble[length] = '\0';
	
	if (!get_params()) { 
		printf("ERROR: Corrupted preamble.\n"); 
		return -1;
	}

	if (Nr_vert > N) {
		printf("Too many vertices! Recompile with N > %d\n", Nr_vert);
		return -1;
	}

	for ( i = 0
		 ; i < Nr_vert && fread(Bitmap[i], 1, (int)((i + 8)/8), fp)
		 ; i++ );

	fclose(fp);

	g->nodes = Nr_vert;

	for (i = 0; i < g->nodes; i++) {
		for (j = 0; j < g->nodes; j++) {
			if (get_edge(i, j))
	       bitmap[(j+1)/CHARBITS][i+1] |= (1 << ((j+1) % CHARBITS));
		}
	}

	return 0;
}

static int get_params(void)
                      /* getting Nr_vert and Nr_edge from the preamble string,
			containing Dimacs format "p ??? num num" */
{
	char c, *tmp;
	char * pp = Preamble;
	int stop = 0;
	tmp = (char *)calloc(100, sizeof(char));
	
	Nr_vert = Nr_edges = 0;
	
	while (!stop && (c = *pp++) != '\0'){
		switch (c)
		  {
			case 'c':
			  while ((c = *pp++) != '\n' && c != '\0');
			  break;
			  
			case 'p':
			  sscanf(pp, "%s %d %d\n", tmp, &Nr_vert, &Nr_edges);
			  stop = 1;
			  break;
			  
			default:
			  break;
		  }
	}
	
	free(tmp);
	
	if (Nr_vert == 0 || Nr_edges == 0)
	  return 0;  /* error */
	else
	  return 1;
	
}

/*************************************************************************/
/**************      PARSE SHYU FILE FORMAT FUNCTION     *****************/
/*************************************************************************/


int parse_shyu(FILE *fp, Graph *g)
{
  char      str[MAXBUF], *ptr = NULL;
  int       fromNode, toNode, nodes;
  int       weight, i;

  /* Number of nodes in the graph  */
  if (fgets(str, MAXBUF, fp) == NULL) {
    fprintf(stderr, "%s: Error parsing file\n", __FUNCTION__);
    return -1;
  }
  sscanf(str, "%d", &nodes);

  /* Parse the weights of the nodes in the graph and add the node */
  /* along with its weight to the graph                           */
  if (fgets(str, MAXBUF, fp) == NULL) {
    fprintf(stderr, "%s: Error parsing file\n", __FUNCTION__);
    return -1;
  }

  ptr = strtok(str, " ");
  for (i = 0; i < nodes; i++) {
    weight = atoi(ptr);
/*
    printf("Adding node %d with weight %f\n", i, (float) weight);
*/
    addWtNodetoGraph(g, i, (float) weight);
    ptr = strtok(NULL, " ");
  }

  /* Parse the adjacency matrix in the file */
  fromNode = 0;
  while (fgets(str, MAXBUF, fp) != NULL) {
    ptr = strtok(str, " ");
    for (toNode = 0; toNode < nodes; toNode++) {
      i = atoi(ptr);
      if (i == 1) {
/*
        printf("Adding edge from %d to %d\n", fromNode, toNode);
*/
        addDirEdgetoGraph(g, fromNode, toNode);
      }
      ptr = strtok(NULL, " ");
    }
    fromNode++;
  }

  return 0;
}


/*************************************************************************/
/**************      PARSE RAKA FILE FORMAT FUNCTION     *****************/
/*************************************************************************/

int parse_raka(FILE *fp, Graph *g)
{
  char      str[MAXBUF], *ptr = NULL;
  int       fromNode, toNode, nodes;
  int       weight, i;

  /* Number of nodes in the graph  */
  if (fgets(str, MAXBUF, fp) == NULL) {
    fprintf(stderr, "%s: Error parsing file\n", __FUNCTION__);
    return -1;
  }
  if (fgets(str, MAXBUF, fp) == NULL) {
    fprintf(stderr, "%s: Error parsing file\n", __FUNCTION__);
    return -1;
  }
  sscanf(str, "%d", &nodes);

  /* Skipping all the position related data and the comment on Weights */
  while ((ptr = fgets(str, MAXBUF, fp)) != NULL) {
    if (str[0] == '*')
      break;
  }
  if (ptr == NULL) {
    fprintf(stderr, "%s: Error parsing file\n", __FUNCTION__);
    return -1;
  }

  /* Parse the weights of the nodes in the graph and add the node */
  /* along with its weight to the graph                           */
  for (i = 0; i < nodes; i++) {
    ptr = fgets(str, MAXBUF, fp);
    if (ptr == NULL) {
      fprintf(stderr, "%s: Error parsing file\n", __FUNCTION__);
      return -1;
    }
    weight = atoi(str);
/*
    printf("Adding node %d with weight %f\n", i, (float) weight);
*/
    addWtNodetoGraph(g, i, (float) weight);
  }

  /* Skip the comment line on CONNECTIONS */
  if (fgets(str, MAXBUF, fp) == NULL) {
    fprintf(stderr, "%s: Error parsing file\n", __FUNCTION__);
    return -1;
  }

  /* Parse the adjacency matrix in the file */
  fromNode = 0;
  while (fgets(str, MAXBUF, fp) != NULL) {
    ptr = strtok(str, " ");
    for (toNode = 0; toNode < nodes; toNode++) {
      i = atoi(ptr);
      if ((toNode != fromNode) && (i == 1)) {
/*
        printf("Adding edge from %d to %d\n", fromNode, toNode);
*/
        addDirEdgetoGraph(g, fromNode, toNode);
      }
      ptr = strtok(NULL, " ");
    }
    fromNode++;
  }

  if (fromNode != nodes) {
    fprintf(stderr, "Error parsing the connections\n");
    return -1;
  }

  return 0;
}

